# Splash scene - first scene the user sees
import pygwidgets
import pyghelpers
from Constants import *

class SceneSplash(pyghelpers.Scene):
    def __init__(self, window):
        self.window = window

        self.backgroundImage = pygwidgets.Image(self.window,
                                                (0, 0), 'images/splashBackground.jpg')
        self.dodgerImage = pygwidgets.Image(self.window,
                                                (150, 30), 'images/dodger.png')
        
        self.startButton = pygwidgets.CustomButton(self.window, (250, 500),
                                                up='images/startNormal.png',
                                                down='images/startDown.png',
                                                over='images/startOver.png',
                                                disabled='images/startDisabled.png',
                                                enterToActivate=True)

        self.quitButton = pygwidgets.CustomButton(self.window, (30, 650),
                                                up='images/quitNormal.png',
                                                down='images/quitDown.png',
                                                over='images/quitOver.png',
                                                disabled='images/quitDisabled.png')

        self.highScoresButton = pygwidgets.CustomButton(self.window, (360, 650),
                                                up='images/gotoHighScoresNormal.png',
                                                down='images/gotoHighScoresDown.png',
                                                over='images/gotoHighScoresOver.png',
                                                disabled='images/gotoHighScoresDisabled.png')

    def getSceneKey(self):
        return SCENE_SPLASH

    def handleInputs(self, events, keyPressedList):
        for event in events:
            if self.startButton.handleEvent(event):
                self.goToScene(SCENE_PLAY)
            elif self.quitButton.handleEvent(event):
                self.quit()
            elif self.highScoresButton.handleEvent(event):
                self.goToScene(SCENE_HIGH_SCORES)

    def draw(self):
        self.backgroundImage.draw()
        self.dodgerImage.draw()
        self.startButton.draw()
        self.quitButton.draw()
        self.highScoresButton.draw()



#  Play scene - the main game play scene
from pygame.locals import *
import pygwidgets
import pyghelpers
from Player import *
from Baddies import *
from Goodies import *

def showCustomYesNoDialog(theWindow, theText):
    oDialogBackground = pygwidgets.Image(theWindow, (40, 250),
                                            'images/dialog.png')
    oPromptDisplayText = pygwidgets.DisplayText(theWindow, (0, 290),
                                            theText, width=WINDOW_WIDTH,
                                            justified='center', fontSize=36)

    oYesButton = pygwidgets.CustomButton(theWindow, (320, 370),
                                            'images/gotoHighScoresNormal.png',
                                            over='images/gotoHighScoresOver.png',
                                            down='images/gotoHighScoresDown.png',
                                            disabled='images/gotoHighScoresDisabled.png')

    oNoButton = pygwidgets.CustomButton(theWindow, (62, 370),
                                            'images/noThanksNormal.png',
                                            over='images/noThanksOver.png',
                                            down='images/noThanksDown.png',
                                            disabled='images/noThanksDisabled.png')

    choiceAsBoolean = pyghelpers.customYesNoDialog(theWindow,
                                            oDialogBackground, oPromptDisplayText,
                                            oYesButton, oNoButton)
    return choiceAsBoolean

BOTTOM_RECT = (0, GAME_HEIGHT + 1, WINDOW_WIDTH,
                                WINDOW_HEIGHT - GAME_HEIGHT)
STATE_WAITING = 'waiting'
STATE_PLAYING = 'playing'
STATE_GAME_OVER = 'game over'

class ScenePlay(pyghelpers.Scene):

    def __init__(self, window):
        self.window = window

        self.controlsBackground = pygwidgets.Image(self.window,
                                        (0, GAME_HEIGHT),
                                        'images/controlsBackground.jpg')

        self.quitButton = pygwidgets.CustomButton(self.window,
                                        (30, GAME_HEIGHT + 90),
                                        up='images/quitNormal.png',
                                        down='images/quitDown.png',
                                        over='images/quitOver.png',
                                        disabled='images/quitDisabled.png')

        self.highScoresButton = pygwidgets.CustomButton(self.window,
                                        (190, GAME_HEIGHT + 90),
                                        up='images/gotoHighScoresNormal.png',
                                        down='images/gotoHighScoresDown.png',
                                        over='images/gotoHighScoresOver.png',
                                        disabled='images/gotoHighScoresDisabled.png')

        self.newGameButton = pygwidgets.CustomButton(self.window,
                                        (450, GAME_HEIGHT + 90),
                                        up='images/startNewNormal.png',
                                        down='images/startNewDown.png',
                                        over='images/startNewOver.png',
                                        disabled='images/startNewDisabled.png',
                                        enterToActivate=True)

        self.soundCheckBox = pygwidgets.TextCheckBox(self.window,
                                        (430, GAME_HEIGHT + 17),
                                        'Background music',
                                        True, textColor=WHITE)

        self.gameOverImage = pygwidgets.Image(self.window, (140, 180),
                                        'images/gameOver.png')

        self.titleText = pygwidgets.DisplayText(self.window,
                                        (70, GAME_HEIGHT + 17),
                                        'Score:                                 High Score:',
                                        fontSize=24, textColor=WHITE)

        self.scoreText = pygwidgets.DisplayText(self.window,
                                        (80, GAME_HEIGHT + 47), '0',
                                        fontSize=36, textColor=WHITE,
                                        justified='right')

        self.highScoreText = pygwidgets.DisplayText(self.window,
                                        (270, GAME_HEIGHT + 47), '',
                                        fontSize=36, textColor=WHITE,
                                        justified='right')

        pygame.mixer.music.load('sounds/background.mid')
        self.dingSound = pygame.mixer.Sound('sounds/ding.wav')
        self.gameOverSound = pygame.mixer.Sound('sounds/gameover.wav')

        # Instantiate objects
        self.oPlayer = Player(self.window)
        self.oBaddieMgr = BaddieMgr(self.window)
        self.oGoodieMgr = GoodieMgr(self.window)

        self.highestHighScore = 0
        self.lowestHighScore = 0
        self.backgroundMusic = True
        self.score = 0
        self.playingState = STATE_WAITING

    def getSceneKey(self):
        return SCENE_PLAY

    def enter(self, data):
        self.getHiAndLowScores()

    def getHiAndLowScores(self):
        # Ask the High Scores scene for a dict of  scores
        # that looks like this:
        #  {'highest': highestScore, 'lowest': lowestScore}
        infoDict = self.request(SCENE_HIGH_SCORES, HIGH_SCORES_DATA)
        self.highestHighScore = infoDict['highest']
        self.highScoreText.setValue(self.highestHighScore)
        self.lowestHighScore = infoDict['lowest']

    def reset(self):   # start a new game
        self.score = 0
        self.scoreText.setValue(self.score)
        self.getHiAndLowScores()

        # Tell the managers to reset themselves
        self.oBaddieMgr.reset()
        self.oGoodieMgr.reset()

        if self.backgroundMusic:
            pygame.mixer.music.play(-1, 0.0)
        self.newGameButton.disable()
        self.highScoresButton.disable()
        self.soundCheckBox.disable()
        self.quitButton.disable()
        pygame.mouse.set_visible(False)

    def handleInputs(self, eventsList, keyPressedList):
        if self.playingState == STATE_PLAYING:
            return  # ignore button events while playing

        for event in eventsList:
            if self.newGameButton.handleEvent(event):
                self.reset()
                self.playingState = STATE_PLAYING

            if self.highScoresButton.handleEvent(event):
                self.goToScene(SCENE_HIGH_SCORES)

            if self.soundCheckBox.handleEvent(event):
                self.backgroundMusic = self.soundCheckBox.getValue()

            if self.quitButton.handleEvent(event):
                self.quit()

    def update(self):
        if self.playingState != STATE_PLAYING:
            return  # only update when playing

        # Move the Player to the mouse position, get back its rect
        mouseX, mouseY = pygame.mouse.get_pos()
        playerRect = self.oPlayer.update(mouseX, mouseY)

        # Tell the GoodieMgr to move all Goodies
        # Returns the number of Goodies that the Player contacted
        nGoodiesHit = self.oGoodieMgr.update(playerRect)
        if nGoodiesHit > 0:
            self.dingSound.play()
            self.score = self.score + (nGoodiesHit * POINTS_FOR_GOODIE)

        # Tell the BaddieMgr to move all the Baddies
        # Returns the number of Baddies that fell off the bottom
        nBaddiesEvaded  = self.oBaddieMgr.update()
        self.score = self.score + (nBaddiesEvaded * POINTS_FOR_BADDIE_EVADED)
        
        self.scoreText.setValue(self.score)

        # Check if the Player has hit any Baddie
        if self.oBaddieMgr.hasPlayerHitBaddie(playerRect):
            pygame.mouse.set_visible(True)
            pygame.mixer.music.stop()

            self.gameOverSound.play()
            self.playingState = STATE_GAME_OVER
            self.draw()  # force drawing of game over message

            if self.score > self.lowestHighScore:
                scoreString = 'Your score: ' + str(self.score) + '\n'
                if self.score > self.highestHighScore:
                    dialogText = (scoreString +
                                       'is a new high score, CONGRATULATIONS!')
                else:
                    dialogText = (scoreString +
                                      'gets you on the high scores list.')

                result = showCustomYesNoDialog(self.window, dialogText)
                if result: # navigate
                    self.goToScene(SCENE_HIGH_SCORES, self.score)  

            self.newGameButton.enable()
            self.highScoresButton.enable()
            self.soundCheckBox.enable()
            self.quitButton.enable()
    
    def draw(self):
        self.window.fill(BLACK)
    
        # Tell the managers to draw all the Baddies and Goodies
        self.oBaddieMgr.draw()
        self.oGoodieMgr.draw()
    
        # Tell the Player to draw itself
        self.oPlayer.draw()
    
        # Draw all the info at the bottom of the window
        self.controlsBackground.draw()
        self.titleText.draw()
        self.scoreText.draw()
        self.highScoreText.draw()
        self.soundCheckBox.draw()
        self.quitButton.draw()
        self.highScoresButton.draw()
        self.newGameButton.draw()

        if self.playingState == STATE_GAME_OVER:
            self.gameOverImage.draw()

    def leave(self):
        pygame.mixer.music.stop()



#  Dodger main program
#
# Instantiates 3 scenes, creates and starts the scene manager
#
#  Original version by Al Sweigart from his book "Invent With Python"
#    (concept, graphics, and sounds used by permission from Al Sweigart)

# 1 - Import packages
import os
# The next line is here just in case you are running from the command line
os.chdir(os.path.dirname(os.path.abspath(__file__)))

import pygame
import pyghelpers
from SceneSplash import *
from ScenePlay import *
from SceneHighScores import *

# 2 - Define constants
FRAMES_PER_SECOND = 40

# 3 - Initialize the world
pygame.init()
window = pygame.display.set_mode((WINDOW_WIDTH, WINDOW_HEIGHT))

# 4 - Load assets: image(s), sounds,  etc.

# 5 - Initialize variables
# Instantiate all scenes and store them in a dict (new approach for pyghelpers 1.1)
scenesList = {SCENE_SPLASH: SceneSplash(window),
                    SCENE_HIGH_SCORES: SceneHighScores(window),
                    SCENE_PLAY: ScenePlay(window)}

# Create the scene manager, passing in the scenes list and the FPS
oSceneMgr = pyghelpers.SceneMgr(scenesList, FRAMES_PER_SECOND)

# Tell the Scene Manager to start running
oSceneMgr.run()



# Player

import pygwidgets
from Constants import *

class Player():
    def __init__(self, window):
        self.window = window
        self.image = pygwidgets.Image(window,
                                (-100, -100), 'images/player.png')
        playerRect = self.image.getRect()
        self.maxX = WINDOW_WIDTH - playerRect.width
        self.maxY = GAME_HEIGHT - playerRect.height

    # Every frame, move the Player icon to the mouse position
    # Limits the x- and y-coordinates to the game area of the window
    def update(self, x, y):
        if x < 0:
            x = 0
        elif x > self.maxX:
            x = self.maxX
        if y < 0:
            y = 0
        elif y > self.maxY:
            y = self.maxY

        self.image.setLoc((x, y))
        return self.image.getRect()

    def draw(self):
        self.image.draw()



# Goodie and GoddieMgr classes
import pygame
import pygwidgets
import random
from Constants import *

class Goodie():
    MIN_SIZE = 10
    MAX_SIZE = 40
    MIN_SPEED = 1
    MAX_SPEED = 8
    # Load the image once
    GOODIE_IMAGE = pygame.image.load('images/goodie.png')
    RIGHT = 'right'
    LEFT = 'left'

    def __init__(self, window):
        self.window = window
        size = random.randrange(Goodie.MIN_SIZE, Goodie.MAX_SIZE + 1)
        self.y = random.randrange(0, GAME_HEIGHT - size)

        self.direction = random.choice([Goodie.LEFT, Goodie.RIGHT])
        if self.direction == Goodie.LEFT:  # start on right side of the window
            self.x = WINDOW_WIDTH
            self.speed = - random.randrange(Goodie.MIN_SPEED,
                                                            Goodie.MAX_SPEED + 1)
            self.minLeft = - size
        else:  # start on left side of the window
            self.x = 0 - size
            self.speed = random.randrange(Goodie.MIN_SPEED,
                                                          Goodie.MAX_SPEED + 1)

        self.image = pygwidgets.Image(self.window,
                                                     (self.x, self.y), Goodie.GOODIE_IMAGE)
        percent = int((size * 100) / Goodie.MAX_SIZE)
        self.image.scale(percent, False)

    def update(self):
        self.x = self.x + self.speed
        self.image.setLoc((self.x, self.y))
        if self.direction == Goodie.LEFT:
            if self.x < self.minLeft:
                return True  # needs to be deleted
            else:
                return False  # stays in window
        else:
            if self.x > WINDOW_WIDTH:
                return True  # needs to be deleted
            else:
                return False  # stays in window

    def draw(self):
        self.image.draw()

    def collide(self, playerRect):
        collidedWithPlayer = self.image.overlaps(playerRect)
        return collidedWithPlayer


class GoodieMgr():
    GOODIE_RATE_LO = 90
    GOODIE_RATE_HI = 111

    def __init__(self, window):
        self.window = window
        self.reset()

    def reset(self):  # Called when starting a new game
        self.goodiesList = []
        self.nFramesTilNextGoodie = GoodieMgr.GOODIE_RATE_HI

    def update(self, thePlayerRect):
        # Tell each Goodie to update itself.
        # If a Goodie goes off an edge, remove it
        # Count up all Goodies that contact the player and remove them.
        nGoodiesHit = 0
        goodiesListCopy = self.goodiesList.copy()
        for oGoodie in goodiesListCopy:
            deleteMe = oGoodie.update()
            if deleteMe:
                self.goodiesList.remove(oGoodie)  # remove this Goodie

            elif oGoodie.collide(thePlayerRect):
                self.goodiesList.remove(oGoodie)  # remove this Goodie
                nGoodiesHit = nGoodiesHit + 1
        
        # If the correct amount of frames have passed,
        # add a new Goodie (and reset the counter)
        self.nFramesTilNextGoodie = self.nFramesTilNextGoodie - 1
        if self.nFramesTilNextGoodie == 0:
            oGoodie = Goodie(self.window)
            self.goodiesList.append(oGoodie)
            self.nFramesTilNextGoodie = random.randrange(
                                                            GoodieMgr.GOODIE_RATE_LO,
                                                            GoodieMgr.GOODIE_RATE_HI)

        return nGoodiesHit  # return number of Goodies that contacted player

    def draw(self):
        for oGoodie in self.goodiesList:
            oGoodie.draw()



"""
pyghelpers is a collection of classes and functions written in Python for use with Pygame.

pyghelpers is pronounced "pig helpers".

Developed by Irv Kalb  -  Irv at furrypants.com

Full documentation at:   https://pyghelpers.readthedocs.io/en/latest/


pyghelpers contains the following classes:

- Timer - a simple timer
- CountUpTimer - a timer that counts up from zero
- CountDownTimer - a timer that counts down from a starting point
- SceneMgr - allows for a Pygame program with multiple scenes
- Scene - base class for a scene managed by the SceneMgr

pyghelpers also contains the following functions:

- textYesNoDialog - a text-based dialog box allowing for one or two answers (yes/no, or just OK)
- customYesNoDialog - a dialog box with custom graphics (yes/no, or just OK)
- textAnswerDialog - a text-based dialog box allowing the user to enter a string
- customAnswerDialog - a dialog box with custom graphics that allows the user to enter a string


While not required, manyhelpers allow the use of a callback (a function or method to be called when an action happens)
    Any widget that uses a callback should be set up like this: 
    def <callbackMethodName>(self, nickName):
    When the appropriate action happens, the callback method will be called and the nickName will be passed.
    If you don't need the nickname, you can just ignore that parameter

"""
"""
************************************************************************************************

Simplified BSD License:

Copyright 2017 Irv Kalb. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Irv Kalb ''AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Irv Kalb OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of Irv Kalb.

******************************************************************************************


History:

7/23 Version 1.1
    SceneMgr: Big change to startup:
       The main program should now create a dictionary of sceneKey: sceneObject pairs
           and get pass this in when creating the SceneMgr.
       This eliminates the earlier need for each scene to have a getSceneKey() method.
       (The old approach is still supported for backward compatibility)
    Scene: Added addScene to add a scene dynamically (pass in a sceneKey and sceneObject)
    Scene: Added removeScene to remove an existing scene to free up memory (pass in a sceneKey)
       (This can be called from the current scene, as long as the next line is a gotoScene)
2/23
    Timer, CountDownTimer, CountUpTimer: checks for and handle multiple pauses and resumes.
    CountDownTimer: modified ended() method to check the time
       Client code doesn't need to call getTime()   Suggested by Lando Chan
    CountUpTimer, CountDownTimer: added forceFullHHMMSS for full HH:MM:SS presentation
10/22
   SceneMgr:  Added optional frame rate display for debugging
4/22
    Timer, CountUpTimer, CountDownTimer: Added pause() and resume()

11/21  Version 1.0.3
    Cleaned up some documentation of parameters
    Removed all FileIO functions
    Added __all__ to define what gets imported when you import *
    Change the SceneMgr so you pass in list of Scenes objects instead of a dictionary.
       Each scene must implement getSceenKey to define the scene key.
       First scene in the list is the starting scene.
    Use ABC for abstract base class and abstract methods
    Updated code using f strings in Timer's getTimeInHHMMSS, and clean up of timers
    Turn off key repeating when going to a new scene
    Changed SceneMgr _goToScene method to goToScene

5/26/20  Version 1.0.2
    Added __version__  and function  getVersion()

Version 1.0     01/13/20

"""

__all__ = [
    'CountDownTimer',
    'CountUpTimer',
    'DIALOG_BACKGROUND_COLOR',
    'DIALOG_BLACK',
    'Scene',
    'SceneMgr',
    'Timer',
    'customAnswerDialog',
    'customYesNoDialog',
    'textAnswerDialog',
    'textYesNoDialog',
]

import pygame
from pygame.locals import *
import pygwidgets
import sys
import time
import os
from abc import ABC, abstractmethod


def getVersion():
    """Returns the current version number of the pyghelpers package"""
    version = "1.1"
    return version



#  Timer
class Timer():
    """
    This class is used to create a very simple Timer.

    Typical use:

    1)  Create a Timer object:

        myTimer = pyghelpers.Timer(10)

    2)  When you want the timer to start running, make this call:

        myTimer.start()

        You can also call this method to restart the timer after it finishes.

    3)  In your big loop, check to see if the timer has finished:

        finished = myTimer.update()

        Normally returns False, but returns True when the timer is finished

    Parameters:
        | timeInSeconds - the duration of the timer, in seconds (integer or float)

    Optional keyword parameters:
        | nickname - an internal name to associate with this timer (defaults to None)
        | callback - a function or object.method to be called back when the timer is finished
        |            The nickname of the timer will be passed in if a callback is made (defaults to None)

    """

    def __init__(self, timeInSeconds, nickname=None, callBack=None):
        self.timeInSeconds = timeInSeconds
        self.nickname = nickname
        self.callBack = callBack
        self.savedSecondsElapsed = 0.0
        self.running = False
        self.paused = False
        self.pauseCounter = 0  # counts how many calls to pause without a resume

    def start(self, newTimeInSeconds=None):
        """Start the timer running (starts at zero).
        Allows you to optionally specify a different amount of time.

        Optional keyword parameter:
        | newTimeInSeconds - a new duration for the timer (integer or float, default None)

        """
        if newTimeInSeconds is not None:
            self.timeInSeconds = newTimeInSeconds
        self.running = True
        self.startTime = time.time()
        self.paused = False
        self.timePaused = None
        self.pauseCounter = 0

    def update(self):
        """Call this in every frame to update the timer

        Returns:
           |   False - most of the time
           |   True - when the timer is finished
           |          (you can use this indication, or set up a callback)

        """
        if not self.running:
            return False
        self.savedSecondsElapsed = time.time() - self.startTime
        if self.savedSecondsElapsed < self.timeInSeconds:
            return False  # running but hasn't reached limit

        else:  # timer has finished
            self.running = False
            if self.callBack is not None:
                self.callBack(self.nickname)

            return True  # True here means that the timer has ended

    def getTime(self):
        """ Call this if you want to know how much has elapsed

        Returns:
           |   0 - if the Timer is not running
           |   seconds elapsed since start, as a float

        """
        if self.running or self.paused:
            self.savedSecondsElapsed = time.time() - self.startTime

        return self.savedSecondsElapsed

    def pause(self):
        """Pauses the timer"""
        self.pauseCounter = self.pauseCounter + 1
        self.timePaused = time.time()
        self.paused = True

    def resume(self):
        """Resumes the timer after a pause"""
        if self.pauseCounter == 0:
            print('Warning - called resume timer but the timer is not paused ... ignored')
        else:
            self.pauseCounter = self.pauseCounter -1
        if self.pauseCounter != 0:
            return  # don't resume

        # OK to resume
        pauseTime = time.time() - self.timePaused
        self.secondsStart = self.secondsStart + pauseTime
        self.paused = False

    def stop(self):
        """Stops the timer"""
        self.getTime()  # remembers final self.savedSecondsElapsed
        self.running = False
        self.paused = False
        self.pauseCounter = 0


# CountUpTimer class
class CountUpTimer():
    """
    This class is used to create a Timer that counts up (starting at zero).

    Its intended use is where you want to continuously display the time in a window (using a DisplayText object).

    Typical use:

    1)  Create a CountUpTimer object:

        myTimer = pyghelpers.CountUpTimer()

    2)  When you want the timer to start running, make this call:

        myTimer.start()

        This method can also be called to restart the timer.

    3)  Whenever you want to get the current time (in seconds since start), you can call any of:

        theTime = myTimer.getTime() # gets time as a float

        theTime = myTimer.getTimeInSeconds() # gets the time as an integer number of seconds

        theTime = myTimer.getTimeInHHMMSS() # gets the time in HH:MM:SS string format

        One of the above should be called every time through your main loop.

    4)  If you want to stop the timer, call:

        myTimer.stop()


    Parameters:
        | none

    """

    def __init__(self):
        self.running = False
        self.savedSecondsElapsed = 0.0
        self.secondsStart = 0  # safeguard
        self.paused = False
        self.timePaused = None
        self.pauseCounter = 0  # counts how many calls to pause without a resume

    def start(self):
        """Start the timer running (starts at zero).  Can be called to restart the timer, for example to play a game multiple times"""
        self.secondsStart = time.time()  # get the current seconds and save the value
        self.running = True
        self.paused = False
        self.savedSecondsElapsed = 0.0
        self.pauseCounter = 0

    def getTime(self):
        """Returns the time elapsed as a float"""
        if not self.running or self.paused:
            return self.savedSecondsElapsed  # do nothing
        
        self.savedSecondsElapsed = time.time() - self.secondsStart
        return self.savedSecondsElapsed  # returns a float

    def getTimeInSeconds(self):
        """Returns the time elapsed as an integer number of seconds"""
        nSeconds = int(self.getTime())
        return nSeconds

    # Updated version using fStrings
    def getTimeInHHMMSS(self, nMillisecondsDigits=0, forceFullHHMMSS= False):
        """Returns the elapsed time as a HH:MM:SS.mmm formatted string

        Parameters:

        Optional keyword parameters:
            | nMillisecondsDigits - number of milliseconds digits to include (defaults to 0)
            |    If specified, returned string will include nMillisecondsDigits of milliseconds digits
            | forceFullHHMMSS forces the output to be in full HH:MM:SS format (defaults to False)

        """
        nSeconds = self.getTime()
        mins, secs = divmod(nSeconds, 60)
        hours, mins = divmod(int(mins), 60)

        if nMillisecondsDigits > 0:
            secondsWidth = nMillisecondsDigits + 3
        else:
            secondsWidth = 2

        if forceFullHHMMSS:
            output = f'{hours:02d}:{mins:02d}:{secs:0{secondsWidth}.{nMillisecondsDigits}f}'
        elif hours > 0:
            output = f'{hours:d}:{mins:02d}:{secs:0{secondsWidth}.{nMillisecondsDigits}f}'
        elif mins > 0:
            output = f'{mins:d}:{secs:0{secondsWidth}.{nMillisecondsDigits}f}'
        else:
            output = f'{secs:.{nMillisecondsDigits}f}'

        return output
    
    def stop(self):
        """Stops the timer"""
        self.getTime()   # remembers final self.savedSecondsElapsed
        self.running = False
        self.paused = False
        self.pauseCounter = 0

    def pause(self):
        """Pauses the timer"""
        self.pauseCounter = self.pauseCounter + 1
        self.timePaused = time.time()
        self.paused = True

    def resume(self):
        """Resumes the timer after a pause"""
        if self.pauseCounter == 0:
            print('Warning - called resume timer but the timer is not paused ... ignored')
        else:
            self.pauseCounter = self.pauseCounter -1
        if self.pauseCounter != 0:
            return  # don't resume

        # OK to resume
        pauseTime = time.time() - self.timePaused
        self.secondsStart = self.secondsStart + pauseTime
        self.paused = False



#
# CountDownTimer class
#
class CountDownTimer():
    """
    This class is used to create a Timer that counts down from a given starting number of seconds.
    Its intended use is where you want to continuously display the time in a window (using a DisplayText object).

    Typical use:

    1)  Create a CountDownTimer object:

        myTimer = pyghelpers.CountDownTimer(60)   # start the timer at 60 seconds

    2)  When you want the timer to start running, make this call:

        myTimer.start()

        This method can also be used to restart the timer.

    3)  Whenever you want to get the remaining time (in seconds since start), you can call any of:

        theTime = myTimer.getTime() # gets time as a float

        theTime = myTimer.getTimeInSeconds() # gets the time as an integer number of seconds

        theTime = myTimer.getTimeInHHMMSS() # gets the time in HH:MM:SS string format

    4)  If you want to stop the timer, call:

        myTimer.stop()


    Parameters:
        | nStartingSeconds - the starting point for the timer, in seconds (integer or float)

    Optional keyword parameters:
        | stopAtZero - should the timer stop when it reaches zero (defaults to True)
        | nickname - an internal name used to refer to this timer (defaults to None)
        | callback - a function or object.method to be called back when the timer is finished
        |            The nickname of the timer will be passed in when the callback is made (defaults to None)

    """

    def __init__(self, nStartingSeconds, stopAtZero=True, nickname=None, callBack=None):
        self.nStartingSeconds = nStartingSeconds
        self.stopAtZero = stopAtZero
        self.nickname = nickname
        self.callBack = callBack

        self.running = False
        self.secondsSavedRemaining = 0.0
        self.reachedZero = False
        self.pauseCounter = 0  # counts how many calls to pause without a resume

    def start(self, newStartingSeconds=None):
        """Start the timer running starting at nStartingSeconds (or optional different setting)"""
        secondsNow = time.time()
        if newStartingSeconds is not None:
            self.nStartingSeconds = newStartingSeconds
        self.secondsEnd = secondsNow + self.nStartingSeconds
        self.reachedZero = False
        self.running = True
        self.paused = False
        self.timePaused = None
        self.pauseCounter = 0

    def getTime(self):
        """Returns the remaining time as a float number of seconds"""
        if not self.running or self.paused:
            return self.secondsSavedRemaining
        
        self.secondsSavedRemaining = self.secondsEnd - time.time()
        if self.stopAtZero and (self.secondsSavedRemaining <= 0):
            self.secondsSavedRemaining = 0.0
            self.running = False
            self.reachedZero = True

        return self.secondsSavedRemaining  # returns a float

    def getTimeInSeconds(self):
        """Returns the remaining time as an integer number of seconds"""
        nSeconds = int(self.getTime())
        return nSeconds

    # Updated version using fStrings
    def getTimeInHHMMSS(self, nMillisecondsDigits=0, forceFullHHMMSS=False):
        """Returns the remaining time as a HH:MM:SS.mmm formatted string

        Parameters:

        Optional keyword parameters:
            | nMillisecondsDigits - number of milliseconds digits to include (defaults to 0)
            |    If specified, returned string will include nMillisecondsDigits of milliseconds digits
            | forceFullHHMMSS forces the output to be in full HH:MM:SS format (defaults to False)

        """
        nSeconds = self.getTime()
        mins, secs = divmod(nSeconds, 60)
        hours, mins = divmod(int(mins), 60)

        if nMillisecondsDigits > 0:
            secondsWidth = nMillisecondsDigits + 3
        else:
            secondsWidth = 2

        if forceFullHHMMSS:
            output = f'{hours:02d}:{mins:02d}:{secs:0{secondsWidth}.{nMillisecondsDigits}f}'
        elif hours > 0:
            output = f'{hours:d}:{mins:02d}:{secs:0{secondsWidth}.{nMillisecondsDigits}f}'
        elif mins > 0:
            output = f'{mins:d}:{secs:0{secondsWidth}.{nMillisecondsDigits}f}'
        else:
            output = f'{secs:.{nMillisecondsDigits}f}'


        self.savedSecondsElapsed = output
        return output


    def stop(self):
        """Stops the timer """
        self.getTime()   # remembers final self.savedSecondsElapsed
        self.running = False
        self.paused = False
        self.pauseCounter = 0
        

    def ended(self):
        """Call to see if the timer has reached zero. Should be called every time through the loop"""
        dontCare = self.getTime()   #  called to  set self.reachedZero
        if self.reachedZero:
            self.reachedZero = False  # reset
            if self.callBack is not None:
                self.callBack(self.nickname)
            return True
        else:
            return False

    def pause(self):
        """Pauses the timer"""
        self.pauseCounter = self.pauseCounter + 1
        self.timePaused = time.time()
        self.paused = True

    def resume(self):
        """Resumes the timer after a pause"""
        if self.pauseCounter == 0:
            print('Warning - called resume timer but the timer is not paused ... ignored')
        else:
            self.pauseCounter = self.pauseCounter -1
        if self.pauseCounter != 0:
            return  # don't resume

        # OK to resume
        pauseTime = time.time() - self.timePaused
        self.secondsStart = self.secondsStart + pauseTime
        self.paused = False


#
# Scene Manager
#
class SceneMgr():
    """SceneMgr (Scene Manager)  allows you to build a program with multiple scenes.

    The SceneMgr manages any number of scenes built as subclasses of the "Scene" class.
    For more details, see the "Scene" class.

    Typical use:

    1) Instantiate as many Scenes as you want:
        |  oScene1 = StartingScene(window)
        |  oScene2 = MainScene(window)
        |  oScene3 = SomeOtherScene(window)

    2) Build a dictionary of these scenes:
        |  myScenesDict = {'sceneKey1': oScene1, 'sceneKey2': oScene2, 'sceneKey3': oScene3}

        The keys are any unique strings that you want to use.

            OLDER APPROACH:  Before pyghelpers 1.1, but still works for backwards compatibility
            Build a list of the scenes:
            | myScenesList = [oScene1, oScene2, oScene3]

    3) Instantiate *one* SceneMgr (a singleton):
        |  oSceneMgr = SceneMgr(myScenesDict, 30) # First scene in the dict is the starting scene

            OLDER APPROACH:  Before pyghelpers 1.1, but still works for backwards compatibility
                oSceneMgr = SceneMgr(myScenesList, 30) # First scene in the list is the starting scene

        You can optionally pass a DisplayText object for showing the frame rate, e.g.
        |  oDebugFrameRate = pygwidgets.DisplayText(window, (0, 0), '', fontSize=24)
        |  oSceneMgr = SceneMgr(scenesDictOrList, FRAMES_PER_SECOND, oDebugFrameRate)

    4) Call the run method to start the SceneMgr running:
        |  oSceneMgr.run()  # First scene in the list is the starting scene


    Parameters:
        | scenesDictOrList - is a dictionary or list that consists of:
        |    {<sceneKey>: <sceneObject>, <sceneKey>: <sceneObject>, ...}
        |    OR older (before pyghelpers 1.1):  [<sceneObject>, <sceneObject>, ...]
        |      where each sceneObject is an object instantiated from a scene class
        |      (For details on Scenes, see the Scene class)
        | fps - is the frames per second at which the program should run

    Based on the concept of a "Scene Manager" by Blake O'Hare of Nerd Paradise (nerdparadise.com)

    """
    def __init__(self, scenesDictOrList, fps, oFrameRateDisplay=None):

        # Newer approach (pyghelpers 1.1), pass in a dictionary of {scene keys: scene objects}
        # (No need to have each scene implement a getSceneKey method.)
        if isinstance(scenesDictOrList, dict):
            self.scenesDict = scenesDictOrList
            keysList = list(self.scenesDict)  # get all the keys
            startingKey = keysList[0]  # first key is the starting scene ley
            self.oCurrentScene = self.scenesDict[startingKey]

        else:  # Older style, we start with a list of scenes
            # Build a dictionary, each entry of which is a scene key : scene object
            # Then we call getSceneKey so each scene can identify itself
            self.scenesDict = {}
            for oScene in scenesDictOrList:
                key = oScene.getSceneKey()  # Each scene must return a unique key to identify itself
                self.scenesDict[key] = oScene
            # The first element in the list is the used as the starting scene
            self.oCurrentScene = scenesDictOrList[0]

        self.framesPerSecond = fps
        self.oFrameRateDisplay = oFrameRateDisplay
        self.showFrameRate = oFrameRateDisplay is not None  # for fast checking in main loop
        self.scenesToRemoveList = []

        # Give each scene a reference back to the SceneMgr.
        # This allows any scene to do a goToScene, request, send,
        # or sendAll, which gets forwarded to the scene manager.
        for key, oScene in self.scenesDict.items():
            oScene._setRefToSceneMgr(self)

    def run(self):
        """
        This method implements the main pygame loop.

        It should typically be called as the last line of your main program.

        It is designed to call a standardized set of methods in the current scene.

        All scenes must implement the following methods (polymorphism):
            |   handleInputs()  # called in every frame
            |   draw()          # called in every frame

        The following methods can be implemented in a scene:
            |   enter()  # called once whenever the scene is entered
            |   update()  # called in every frame
            |   leave()  # called once whenever the scene is left

        If any is not implemented, then the version in the Scene base class,
        which only performs a pass statement, will be used.

        """
        clock = pygame.time.Clock()

        # 6 - Loop forever
        while True:
            # If there are any scenes to be removed (keys added by removeScene method)
            if not(self.scenesToRemoveList is []):
                for key in self.scenesToRemoveList:
                    del self.scenesDict[key]
                self.scenesToRemoveList = []  # reset

            keysDownList = pygame.key.get_pressed()

            # 7 - Check for and handle events
            eventsList = []
            for event in pygame.event.get():
                if (event.type == pygame.QUIT) or \
                        ((event.type == pygame.KEYDOWN) and
                        (event.key == pygame.K_ESCAPE)):
                    # Tell current scene we're leaving
                    self.oCurrentScene.leave()  
                    pygame.quit()
                    sys.exit()

                eventsList.append(event)

            # Here, we let the current scene process all events by calling its handleInputs() method
            # do any "per frame" actions in its update() method,
            # and call its draw() method so it can draw everything that needs to be drawn.
            self.oCurrentScene.handleInputs(eventsList, keysDownList)
            self.oCurrentScene.update()
            self.oCurrentScene.draw()
            if self.showFrameRate:
                fps = str(clock.get_fps())
                self.oFrameRateDisplay.setText('FPS: ' + fps)
                self.oFrameRateDisplay.draw()

            # 11 - Update the window
            pygame.display.update()

            # 12 - Slow things down a bit
            clock.tick(self.framesPerSecond)
            

    def _goToScene(self, nextSceneKey, dataForNextScene):
        """Called by a Scene, tells the SceneMgr to go to another scene

        (From the Scene's point of view, it just needs to call its own goToScene method)
        This method:
        - Tells the current scene that it is leaving, calls leave method
        - Gets any data the leaving scene wants to send to the new scene
        - Tells the new scene that it is entering, calls enter method

        Raises:
        - KeyError if the nextSceneKey is not valid

        """
        if nextSceneKey is None:  # meaning, exit
            pygame.quit()
            sys.exit()

        # Call the leave method of the old scene to allow it to clean up.
        # Set the new scene (based on the key) and
        # call the enter method of the new scene.
        self.oCurrentScene.leave()
        pygame.key.set_repeat(0) # turn off repeating characters
        try:
            self.oCurrentScene = self.scenesDict[nextSceneKey]
        except KeyError:
            raise KeyError("Trying to go to scene '" + nextSceneKey +
                "' but that key is not in the dictionary of scenes.")
        self.oCurrentScene.enter(dataForNextScene)


    def _request_respond(self, targetSceneKey, requestID):
        """Internal method, called by a Scene tells SceneMgr to query another scene for information.

        (From the Scene's point of view, it just needs to call its own request method)
        The target scene must implement a method named "respond"

        """
        oTargetScene = self.scenesDict[targetSceneKey]
        info = oTargetScene.respond(requestID)
        return info

    def _send_receive(self, targetSceneKey, sendID, info):
        """Internal method, called by a Scene, tells the Scene Manager to send information to another scene

        (From the sending scene's point of view, it just needs to call its own send method)
        The target scene must implement a method named "receive"

        """
        oTargetScene = self.scenesDict[targetSceneKey]
        oTargetScene.receive(sendID, info)

    def _sendAll_receive(self, oSenderScene, sendID, info):
        """Internal method, called by a Scene tells the Scene Manager to send information to all scenes (other than itself)

        (From the sending scene's point of view, it just needs to call its own sendAll method)
        All scenes must implement a method named "receive"

        """
        for sceneKey in self.scenesDict:
            oTargetScene = self.scenesDict[sceneKey]
            if oTargetScene != oSenderScene:
                oTargetScene.receive(sendID, info)

    def _addScene(self, sceneKey, oNewScene):
        """Called in a Scene, tells the SceneMgr to add a new scene dynamically
        (From the Scene's point of view, it just needs to call its own addScene method)

        The scene must first instantiate a new Scene object, then call this method

        Parameters:
            | sceneKey - an key that uniquely identifies this scene (typically a string)
            | oNewScene - an object of an instance of the new scene

        Raises:
        |    KeyError - if a scene with the same key already exists (must be unique)

        """
        # Ask the new scene for its scene key, and add it to the scenes dictionary
        newSceneKey = oNewScene.getSceneKey()
        if newSceneKey in self.scenesDict:
            raise KeyError('Trying to add a scene with key' + newSceneKey + 'but that scene key already exists')
        self.scenesDict[newSceneKey] = oNewScene
        # Send the new scene a reference to the SceneMgr
        oNewScene._setRefToSceneMgr(self)

    def _removeScene(self, sceneKeyToRemove):
        """Called by a Scene, tells the SceneMgr to remove an existing scene (to free up memory)
        (From the Scene's point of view, it just needs to call its own remove method)

        Parameter:
            | sceneKeyToRemove - the scene key of the scene to be eliminated

        Raises:
            | KeyError if the nextSceneKey is not valid

        """
        if not(sceneKeyToRemove in self.scenesDict):
            raise KeyError('Attempting to remove scene with key ' + sceneKeyToRemove +
                           ' but no scene with that key currently exists (either never defined or removed).')
        # Add the key to a list of keys to be removed
        # The scene(s) will be removed the next time around the main loop.from
        # This allows a scene to make a call to removeScene to remove itself right before doing a goToScene
        #    (and the call will return successfully)
        self.scenesToRemoveList.append(sceneKeyToRemove)



class Scene(ABC):
    """
    The Scene class is an abstract class to be used as a base class for any scenes that you want to create.

    In the startup code for your  application, you create an instance of each of the
    scenes that you want to be available.  Then you build a dictionary like this

        |    {<sceneKey1>:<sceneObject1>, <sceneKey2>:<sceneObject2>, ...}

    Then you pass this dictionary when you instantiate the Scene Manager.
    See the SceneMgr documentation for details.

    To create a scene, you must write a subclass that inherits from this class.
    Your class must implement these methods

        |   __init__(), handleInputs(), draw()

    You will typically overwrite the update() method to do any processing in every frame.

    Other methods can be overridden if you wish:

        | enter(), leave()

    You can add any additional methods that your class requires.
    Additionally, there are other methods to allow you to get or set info, or navigate to other scenes:

        | goToScene(), quit(), request(), respond() and more.

    In the __init__() method of your scene subclass, you will receive a window reference.
    You should copy this into an instance variable like this, and use it when drawing:

        |    def __init__(self, window):
        |        self.window = window
        |        # Add any initialization you want to do here.

    Before version 1.1 of pyghelpers you had to do the following, but
    as of version 1.1, this method is no longer needed.  Instead, you now
    set the scene keys at start up  (much more betterer!).

                OLD: You also need to write a getSceneKey() method that returns a string
                or constant that uniquely identifies the scene.  It is recommended that you
                build and import a Constants.py file that contains constants for each scene,
                and use the key associated with the current scene here.
                |    def getSceneKey(self):
                |        return <string or CONSTANT that identifies this scene>

    When your scene is active, the SceneManager calls a standard set of methods in the current scene.
    Therefore, all scenes must implement these methods (polymorphism):

        |    handleInputs()  # called in every frame
        |    draw()          # called in every frame

    The following methods can optionally be implemented in a scene.  If they are not
    implemented, then the null version in the Scene subclass will be used.
    (The Scene class' default versions only execute a pass statement):

        |    enter()          # called once whenever the scene is entered
        |    update()       # called in every frame
        |    leave()          # called once whenever the scene is left

    When you want to go to a new scene, call:

        |    self.goToScene() and pass in the sceneKey of the scene you want to go to,
        |    and optionally, pass any data you want the next scene to receive in its enter() method.

    If you want to quit the program from your scene, call:

        |    self.quit()

    """
    def __del__(self):
        """Internal method, called when the scene is about to die."""
        self.oSceneMgr = None  # eliminate the reference to the SceneMgr

    def _setRefToSceneMgr(self, oSceneMgr):
        """Internal method to save  a reference to the SceneMgr object

        This exists so each class built from this base class can call methods in the Scene Manager
        That reference is used by the goToScene(), request(), and send ()methods in each Scene
        Do not change or override this method.

        """
        self.oSceneMgr = oSceneMgr

    def enter(self, data):
        """This method is called whenever the user enters a scene.

        Should be overridden if you expect data when your scene is entered.
        Add any code you need to start or re-start the scene

        Parameter:
            |    data - can be of any type agreed to by the old and new scenes

        """
        pass

    # This is no longer needed (as of pyghelpers 1.1)
    # Left in for backwards compatibility, but no longer needs to be an abstract method
    #@abstractmethod
    def getSceneKey(self):
        """This method must return the scene key for this scene
        """
        raise NotImplementedError

    @abstractmethod
    def handleInputs(self, events, keyPressedList):
        """This method is called in every frame of the scene to handle events and key presses

        Your code MUST override this method.

        Parameters:
            |    events - a list of events your method should handle.
            |    keyPressedList - a list of keys that are pressed (a Boolean for each key).

        """
        raise NotImplementedError

    def update(self):
        """This method is called in every frame of the scene do any processing you need to do here

        Your code will typically override this method.

        """
        pass

    @abstractmethod
    def draw(self):
        """This method is called in every frame of the scene to draw anything that needs to be drawn

        Your code MUST override this method.

        """
        raise NotImplementedError

    def leave(self):
        """This method is called whenever the user leaves a scene

        Override this method, and add any code you need to clean up the scene before leaving

        """
        pass

    def quit(self):
        """Call this method if you want to quit, from inside a scene"""
        self.goToScene(None)


    def goToScene(self, nextSceneKey, data=None):
        """Call this method whenever you want to go to a new scene

        Parameters:
            |    nextSceneKey - the scene key (string) of the scene to go to

        Optional keyword parameter:
            |    data - any data you want sent to the next scene (defaults to None)
            |          (The data can be a single value, a list, dictionary, object, etc.)

        """
        self.oSceneMgr._goToScene(nextSceneKey, data)


    def request(self, targetSceneKey, requestID):
        """Call this method to get information from another scene

        The target scene must implement a method named: respond,
        it can return any info in any way the two scenes agree upon

        Parameters:
            |    targetSceneKey - the scene key (string) of the scene to ask for data
            |    requestID - the data you want from the target scene (typically a string)

        """
        info = self.oSceneMgr._request_respond(targetSceneKey, requestID)
        return info

    def send(self, targetSceneKey, sendID, info):
        """Call this method to send information to  another scene

        The other scene must implement a method named:  receive().
        You can pass any info the two scenes agree upon

        Parameters:
            |    targetSceneKey - the scene key (string) of the scene to ask for data
            |    sendID - the type of data you are sending the target scene (typically a string)
            |    info - the actual data to send (can be any type)

        """
        self.oSceneMgr._send_receive(targetSceneKey, sendID, info)

    def sendAll(self, sendID, info):
        """Call this method to send information to all other scenes

        The other scenes must implement a method named:  receive().
        You can pass any info that the sender and all other scenes agree upon

        Parameters:
            |    sendID - the type of data you are sending the target scene (typically a string)
            |    info - the actual data to send (can be any type)

        """
        self.oSceneMgr._sendAll_receive(self, sendID, info)  # pass in self to identify sender

    def respond(self, requestID):
        """Respond to a request for information from some other scene

        You must override this method if your scene expects to handle
        requests for information from other scenes via calls to:  request()

        Parameters:
            |    requestID - identifier of what data to be sent back to the caller

        """
        raise NotImplementedError

    def receive(self, receiveID, info):
        """Receives information from another scene.

        You must override this method if your scene expects to receive information from
        other scenes sending information via calls to:  send

        Parameters:
            |    receiveID - an identifier for what type of information is being received
            |    info - the information sent from another scene

        """
        raise NotImplementedError

    def addScene(self, sceneKey, oScene):
        """Call this method whenever you want to add a new scene dynamically.
        For example, you could have a game with many levels (each implemented as a scene),
        but only have the current level loaded.  As the player completes a level, you
        could do a removeScene on the current level and do an addScene on the
        next level, then do a goToScene on the new level.

        Parameters:
            |    sceneKey - a key to uniquely identify this scene (typically a string)
            |    oScene - an instance of the new scene to be added
            |         (typically, you would instantiate the new scene, and pass in that reference to this call)

        """
        self.oSceneMgr._addScene(sceneKey, oScene)

    def removeScene(self, sceneKey):
        """Call this method whenever you want to remove an existing scene
        You can remove a scene to save memory - the scene object will be deleted.
        The SceneMgr delays removing the scene until the next time thru the main loop.
        Therefore, it is safe to call to removeScene from its own scene, if you immediately
        go to another scene.

        Parameters:
            |    sceneKey - the scene key (string) of the scene to remove

        """
        self.oSceneMgr._removeScene(sceneKey)


#
#  Dialog functions
#
DIALOG_BACKGROUND_COLOR = (0, 200, 200)
DIALOG_BLACK = (0, 0, 0)

def textYesNoDialog(theWindow, theRect, prompt, yesButtonText='Yes', 
                    noButtonText='No', backgroundColor=DIALOG_BACKGROUND_COLOR,
                    textColor=DIALOG_BLACK):
    """A function that puts up a text-based two-button modal dialog (typically Yes/No or OK/Cancel)

    It can also be used to put up a single button alert dialog (typically with an OK button)

    Parameters:
        |    theWindow - the window to draw in
        |    theRect - the rectangle (or tuple) of the dialog box in the application window
        |    prompt - prompt (title) string to be displayed in the dialog box

    Optional keyword parameters:
        |    yesButtonText - text on the Yes button (defaults to 'Yes')
        |    noButtonText - text on the No button (defaults to 'No')
        |       Note:  If noButtonText is None, the nothing will be drawn for the No button
        |              This way, you can present an "alert" box with only an 'OK' button
        |    backgroundColor - rgb background color for the dialog box (defaults to (0, 200, 200))
        |    textColor - rgb color for the prompt text (defaults to black)

    Returns:
        |    True - meaning the Yes button was pressed
        |        or
        |    False - meaning the No button was pressed
        |
        |   (With an alert dialog, you can ignore the returned value, as it will always be True.)

    """
    dialogLeft = theRect[0]
    dialogTop = theRect[1]
    dialogWidth = theRect[2]
    dialogHeight = theRect[3]
    frameRect = pygame.Rect(dialogLeft + 1, dialogTop + 1, dialogWidth - 2, dialogHeight - 2)
    INSET = 30 # inset buttons from the edges of the dialog box

    promptText = pygwidgets.DisplayText(theWindow, (dialogLeft, dialogTop + 30), prompt,
                                        fontSize=24, width=dialogWidth, justified='center', textColor=textColor)

    # Create buttons, fix locations after finding out the size of the button(s)
    showNoButton = not (noButtonText is None)
    if showNoButton:
        noButton = pygwidgets.TextButton(theWindow, (0, 0), noButtonText)
    yesButton = pygwidgets.TextButton(theWindow, (0, 0), yesButtonText)

    yesButtonRect = yesButton.getRect()
    yesButtonHeight = yesButtonRect[3]
    yesButtonWidth = yesButtonRect[2]  # get width
    xPos = dialogLeft + dialogWidth - yesButtonWidth - INSET
    buttonsY = dialogTop + dialogHeight - yesButtonHeight - 20
    if showNoButton:
        noButton.setLoc((dialogLeft + INSET, buttonsY))
    yesButton.setLoc((xPos, buttonsY))

    #print('In dialogYesNo')
    #print('theRect is', theRect)
    #print('frameRect is', frameRect)

    # 6 - Loop forever
    while True:

        # 7 - Check for and handle events
        for event in pygame.event.get():
            if (event.type == QUIT) or \
                ((event.type == KEYDOWN) and (event.key == K_ESCAPE)):
                pygame.quit()
                sys.exit()

            if showNoButton:
                if noButton.handleEvent(event):
                    return False

            if yesButton.handleEvent(event):
                return True

        # 8 - Do any "per frame" actions

        # 9 - Clear the window area before drawing it again
        pygame.draw.rect(theWindow, backgroundColor, theRect)
        pygame.draw.rect(theWindow, DIALOG_BLACK, frameRect, 1)

        # 10 - Draw the window elements
        promptText.draw()
        if showNoButton:
            noButton.draw()
        yesButton.draw()

        # 11 - Update the window
        pygame.display.update()

        # 12 - Slow things down a bit
        #clock.tick(FRAMES_PER_SECOND)  # no need for this


def customYesNoDialog(theWindow, oDialogImage, oPromptText, oYesButton, oNoButton=None):
    """A function that puts up a custom two-button modal dialog (typically Yes/No or OK/Cancel)

    It can also be used to put up a single button alert dialog (with a typcial OK button)

    Parameters:
        |    theWindow - the window to draw in
        |    oDialogImage - an Image object (from pygwidgets) with the background of the dialog box
        |    oPromptText - a TextDisplay object (from pygwidgets) containing the prompt to display
        |    oYesButton - a CustomButton object (from pygwidgets) representing Yes or OK, etc.

    Optional keyword parameter:
        |    oNoButton - a CustomButton object (from pygwidgets) representing No or Cancel, etc. (default None)
        |       Note:  If oNoButton is None, the No button will not be drawn
        |              This way, you can present an "alert" box with only a single button, like 'OK'

    Returns:
        |    True - meaning the Yes button was pressed
        |        or
        |    False - meaning the No button was pressed
        |
        |   (With an alert dialog, you can ignore the returned value, as it will always be True.)

    """

    showNoButton = not (oNoButton is None)

    # 6 - Loop forever
    while True:

        # 7 - Check for and handle events
        for event in pygame.event.get():
            if (event.type == QUIT) or \
                ((event.type == KEYDOWN) and (event.key == K_ESCAPE)):
                pygame.quit()
                sys.exit()

            if showNoButton:
                if oNoButton.handleEvent(event):
                    return False

            if oYesButton.handleEvent(event):
                return True

        # 8 - Do any "per frame" actions

        # 9 - Clear the window area before drawing it again

        # 10 - Draw the window elements
        oDialogImage.draw()
        oPromptText.draw()
        if showNoButton:
            oNoButton.draw()
        oYesButton.draw()

        # 11 - Update the window
        pygame.display.update()

        # 12 - Slow things down a bit
        #clock.tick(FRAMES_PER_SECOND)  # no need for this


def textAnswerDialog(theWindow, theRect, prompt, okButtonText='OK',
                    cancelButtonText='Cancel', backgroundColor=DIALOG_BACKGROUND_COLOR,
                    promptTextColor=DIALOG_BLACK, inputTextColor=DIALOG_BLACK):
    """A function that puts up a text-based two-button answerable modal dialog (typically OK/Cancel)

    Parameters:
        |    theWindow - the window to draw in
        |    theRect - the rectangle (or tuple) of the dialog box in the application window
        |    prompt - prompt (title) string to be displayed in the dialog box

    Optional keyword parameters:
        |    okButtonText - text on the OK button (defaults to 'OK')
        |    cancelButtonText - text on the Cancel button (defaults to 'Cancel')
        |    backgroundColor - rgb background color for the dialog box (defaults to (0, 200, 200))
        |    promptTextColor - rgb color of the prompt text (defaults to black)
        |    inputTextColor - rgb color of the input text (defaults to black)

    Returns:
         |   userAnswer - If user presses OK, returns the text the user typed. Otherwise returns None

    """

    dialogLeft = theRect[0]
    dialogTop = theRect[1]
    dialogWidth = theRect[2]
    dialogHeight = theRect[3]
    INSET = 30 # inset buttons from the edges of the dialog box

    promptText = pygwidgets.DisplayText(theWindow, (dialogLeft, dialogTop + 30), prompt,
                                        fontSize=24, width=dialogWidth, justified='center',
                                        textColor=promptTextColor)

    inputWidth = dialogWidth - (2 * INSET)
    inputText = pygwidgets.InputText(theWindow, (dialogLeft + INSET, dialogTop + 80),
                                     width=inputWidth, initialFocus=True, textColor=inputTextColor)

    cancelButton = pygwidgets.TextButton(theWindow, (0, 0), cancelButtonText)
    okButton = pygwidgets.TextButton(theWindow, (0, 0), okButtonText)

    okButtonRect = okButton.getRect()
    okButtonHeight = okButtonRect[3]
    okButtonWidth = okButtonRect[2]  # get width
    xPos = dialogLeft + dialogWidth - okButtonWidth - INSET
    buttonsY = dialogTop + dialogHeight - okButtonHeight - 20
    cancelButton.setLoc((dialogLeft + INSET, buttonsY))
    okButton.setLoc((xPos, buttonsY))

    # 6 - Loop forever
    while True:

        # 7 - Check for and handle events
        for event in pygame.event.get():
            if (event.type == QUIT) or \
                ((event.type == KEYDOWN) and (event.key == K_ESCAPE)):
                pygame.quit()
                sys.exit()

            if inputText.handleEvent(event) or okButton.handleEvent(event):
                theAnswer = inputText.getValue()
                return theAnswer

            if cancelButton.handleEvent(event):
                return None

        # 8 - Do any "per frame" actions

        # 9 - Clear the window area before drawing it again
        pygame.draw.rect(theWindow, backgroundColor, theRect)
        pygame.draw.rect(theWindow, DIALOG_BLACK, theRect, 1)

        # 10 - Draw the window elements
        promptText.draw()
        inputText.draw()
        cancelButton.draw()
        okButton.draw()

        # 11 - Update the window
        pygame.display.update()

        # 12 - Slow things down a bit
        #clock.tick(FRAMES_PER_SECOND)  # no need for this


def customAnswerDialog(theWindow, oDialogImage, oPromptText, oAnswerText, oOKButton, oCancelButton):
    """A function that puts up a custom two-button modal dialog (typically Yes/No or OK/Cancel)

    Parameters:
        |    theWindow - the window to draw in
        |    oDialogImage - an Image object (from pygwidgets) containing the background of the dialog box
        |    oPromptText - a TextDisplay object (from pygwidgets) containing the prompt to display
        |    oAnswerText - an InputText object (from pygwidgets) where the user types their answer
        |    oOKButton - a CustomButton object (from pygwidgets) representing OK, etc.
        |    oCancelButton - a CustomButton object (from pygwidgets) representing Cancel, etc.

    Returns:
         |    userAnswer - If user presse OK, returns the text the user typed. Otherwise returns None

    """

    # 6 - Loop forever
    while True:

        # 7 - Check for and handle events
        for event in pygame.event.get():
            if (event.type == QUIT) or \
                ((event.type == KEYDOWN) and (event.key == K_ESCAPE)):
                pygame.quit()
                sys.exit()

            if oAnswerText.handleEvent(event) or oOKButton.handleEvent(event):
                userResponse = oAnswerText.getValue()
                return userResponse

            if oCancelButton.handleEvent(event):
                return None

        # 8 - Do any "per frame" actions

        # 9 - Clear the window area before drawing it again

        # 10 - Draw the window elements
        oDialogImage.draw()
        oAnswerText.draw()
        oPromptText.draw()
        oCancelButton.draw()
        oOKButton.draw()

        # 11 - Update the window
        pygame.display.update()

        # 12 - Slow things down a bit
        #clock.tick(FRAMES_PER_SECOND)  # no need for this



# Testing program for 6 dialog boxes
#  By Irv Kalb

# 1 - Import packages
import os
import sys
# The next line is here just in case you are running from the command line
os.chdir(os.path.dirname(os.path.abspath(__file__)))
import pygame
import pygwidgets
import pyghelpers

# 2 - Define constants
BLACK = (0, 0, 0)
BACKGROUND_COLOR = (0, 138, 138)
WHITE = (255, 255, 255)
WINDOW_WIDTH = 640
WINDOW_HEIGHT = 480
FRAMES_PER_SECOND = 30



def showCustomAlertDialog(theWindow, theText):
    oDialogBackground = pygwidgets.Image(theWindow, (60, 120), 'images/dialog.png')
    oPromptDisplayText = pygwidgets.DisplayText(theWindow, (0, 170), theText,
                                width=WINDOW_WIDTH, justified='center', fontSize=36)
    oOKButton = pygwidgets.CustomButton(theWindow, (355, 265),
                                        'images/okNormal.png',
                                        over='images/okOver.png',
                                        down='images/okDown.png',
                                        disabled='images/okDisabled.png')
    userAnswer = pyghelpers.customYesNoDialog(theWindow, oDialogBackground,
                                    oPromptDisplayText, oOKButton, None)
    return userAnswer


def showCustomYesNoDialog(theWindow, theText):
    oDialogBackground = pygwidgets.Image(theWindow, (60, 120), 'images/dialog.png')
    oPromptDisplayText = pygwidgets.DisplayText(theWindow, (0, 170), theText,
                                width=WINDOW_WIDTH, justified='center', fontSize=36)
    oNoButton = pygwidgets.CustomButton(theWindow, (95, 265),
                                        'images/noNormal.png',
                                        over='images/noOver.png',
                                        down='images/noDown.png',
                                        disabled='images/noDisabled.png')
    oYesButton = pygwidgets.CustomButton(theWindow, (355, 265),
                                        'images/yesNormal.png',
                                        over='images/yesOver.png',
                                        down='images/yesDown.png',
                                        disabled='images/yesDisabled.png')
    userAnswer = pyghelpers.customYesNoDialog(theWindow, oDialogBackground,
                                    oPromptDisplayText, oYesButton, oNoButton)
    return userAnswer

def showCustomAnswerDialog(theWindow, theText):
    oDialogBackground = pygwidgets.Image(theWindow, (60, 120), 'images/dialog.png')
    oPromptDisplayText = pygwidgets.DisplayText(theWindow, (0, 170), theText,
                                width=WINDOW_WIDTH, justified='center', fontSize=36)
    oUserInputText = pygwidgets.InputText(theWindow, (225, 220), '',
                                            fontSize=36, initialFocus=True)
    oNoButton = pygwidgets.CustomButton(theWindow, (105, 265),
                                        'images/cancelNormal.png',
                                        over='images/cancelOver.png',
                                        down='images/cancelDown.png',
                                        disabled='images/cancelDisabled.png')
    oYesButton = pygwidgets.CustomButton(theWindow, (375, 265),
                                        'images/okNormal.png',
                                        over='images/okOver.png',
                                        down='images/okDown.png',
                                        disabled='images/okDisabled.png')
    userAnswer = pyghelpers.customAnswerDialog(theWindow, oDialogBackground,
                                    oPromptDisplayText, oUserInputText, oYesButton, oNoButton)
    return userAnswer


# 3 - Initialize the world
pygame.init()
window = pygame.display.set_mode((WINDOW_WIDTH, WINDOW_HEIGHT))
clock = pygame.time.Clock()  # set the speed (frames per second)

# 4 - Load assets: image(s), sounds,  etc.


# 5 - Initialize variables

textAlertButton = pygwidgets.TextButton(window, (75, 370), 'Text Alert')
customAlertButton = pygwidgets.TextButton(window, (75, 420), 'Custom Alert')
textYesNoButton = pygwidgets.TextButton(window, (280, 370), 'Text Yes/No')
customYesNoButton = pygwidgets.TextButton(window, (280, 420), 'Custom Yes/No')
textAnswerButton = pygwidgets.TextButton(window, (485, 370), 'Text Answer')
customAnswerButton = pygwidgets.TextButton(window, (485, 420), 'Custom Answer')

title = pygwidgets.DisplayText(window, (150, 50), 'Click all buttons to test dialogs',
                                    fontSize=36, textColor=WHITE)


# 6 - Loop forever
while True:

    # 7 -  Check for and handle events
    for event in pygame.event.get():
        # check if the event is the X button 
        if event.type == pygame.QUIT:
            # if it is quit the game
            pygame.quit()
            sys.exit()

        if textAlertButton.handleEvent(event):
            # Note, only an OK button, so need to check the value returned from this call (will be True)
            pyghelpers.textYesNoDialog(window, (75, 100, 500, 150),
                                                  'This is an alert!', 'OK', None)
            print('User clicked the OK button')

        if customAlertButton.handleEvent(event):
            # Note, only an OK button, so need to check the value returned from this call (will be True)
            showCustomAlertDialog(window, 'This is an alert!')
            print('User clicked the OK button')

        if textYesNoButton.handleEvent(event):
            returnedValue = pyghelpers.textYesNoDialog(window, (75, 100, 500, 150),
                                                  'Do you want fries with that?', 'Yes', 'No')
            print('Returned value was:', returnedValue)
            if returnedValue:
                print('User clicked Yes')
            else:
                print('User clicked No')

        if customYesNoButton.handleEvent(event):
            returnedValue = showCustomYesNoDialog(window, 'Do you want fries with that?')
            print('Returned value was:', returnedValue)                                             
            if returnedValue:
                print('User clicked Yes')
            else:
                print('User clicked No')

        if textAnswerButton.handleEvent(event):
            returnedValue, userAnswer = pyghelpers.textAnswerDialog(window, (75, 100, 500, 200),
                                                  'What is your favorite flavor of ice cream?', 'OK', 'Cancel')
            print('Returned values were:', returnedValue, userAnswer)                                            
            if returnedValue:
                print('User clicked OK')
                print('Users answer was:', userAnswer)
            else:
                print('User clicked cancel')

        if customAnswerButton.handleEvent(event):
            userAnswer = showCustomAnswerDialog(window,
                                                           'What is your favorite flavor of ice cream?')
            if userAnswer is None:
                print('User clicked Cancel')
            else:
                print('Users answer was:', userAnswer)




    # 8 - Do any "per frame" actions

    # 9 - Clear the screen before drawing it again
    window.fill(BACKGROUND_COLOR)
                          
    # 10 - Draw the screen elements
    title.draw()

    textAlertButton.draw()
    customAlertButton.draw()
    textYesNoButton.draw()
    customYesNoButton.draw()
    textAnswerButton.draw()
    customAnswerButton.draw()
    
    # 11 - Update the screen
    pygame.display.update()

    # 12 - Slow things down a bit
    clock.tick(FRAMES_PER_SECOND)  # make PyGame wait the correct amount







# Baddie and BaddieMgr classes

import pygame
import pygwidgets
import random
from Constants import *

# Baddie class
class Baddie():
    MIN_SIZE = 10
    MAX_SIZE = 40
    MIN_SPEED = 1
    MAX_SPEED = 8
    # Load the image only once
    BADDIE_IMAGE = pygame.image.load('images/baddie.png')

    def __init__(self, window):
        self.window = window
        # Create the image object
        size = random.randrange(Baddie.MIN_SIZE, Baddie.MAX_SIZE + 1)
        self.x = random.randrange(0, WINDOW_WIDTH - size)
        self.y = 0 - size # start above the window
        self.image = pygwidgets.Image(self.window, (self.x, self.y),
                                      Baddie.BADDIE_IMAGE)

        # Scale it
        percent = (size * 100) / Baddie.MAX_SIZE
        self.image.scale(percent, False)
        self.speed = random.randrange(Baddie.MIN_SPEED,
                                                      Baddie.MAX_SPEED + 1)

    def update(self):  # move the Baddie down
        self.y = self.y + self.speed
        self.image.setLoc((self.x, self.y))
        if self.y > GAME_HEIGHT:
            return True  # needs to be deleted
        else:
            return False  # stays in the window

    def draw(self):
        self.image.draw()

    def collide(self, playerRect):
        collidedWithPlayer = self.image.overlaps(playerRect)
        return collidedWithPlayer

# BaddieMgr class
class BaddieMgr():
    ADD_NEW_BADDIE_RATE = 8  # how often to add a new Baddie

    def __init__(self, window):
        self.window = window
        self.reset()

    def reset(self):  # called when starting a new game
        self.baddiesList = []
        self.nFramesTilNextBaddie = BaddieMgr.ADD_NEW_BADDIE_RATE

    def update(self):
        # Tell each Baddie to update itself
        # Count how many Baddies have fallen off the bottom.
        nBaddiesRemoved = 0
        baddiesListCopy = self.baddiesList.copy()
        for oBaddie in baddiesListCopy:
            deleteMe = oBaddie.update()
            if deleteMe:
                self.baddiesList.remove(oBaddie)
                nBaddiesRemoved = nBaddiesRemoved + 1

        # Check if it's time to add a new Baddie
        self.nFramesTilNextBaddie = self.nFramesTilNextBaddie - 1
        if self.nFramesTilNextBaddie == 0:
            oBaddie = Baddie(self.window)
            self.baddiesList.append(oBaddie)
            self.nFramesTilNextBaddie = BaddieMgr.ADD_NEW_BADDIE_RATE

        # Return that count of Baddies that were removed
        return nBaddiesRemoved

    def draw(self):
        for oBaddie in self.baddiesList:
            oBaddie.draw()

    def hasPlayerHitBaddie(self, playerRect):
        for oBaddie in self.baddiesList:
            if oBaddie.collide(playerRect):
                return True
        return False



from setuptools import find_packages, setup

setup(
    name='pyghelpers',
    version='1.0.43',
    author='Irv Kalb',
    author_email='Irv@furrypants.com',
    description='Classes and functions for use with Pygame',
    long_description='A collection of classes and functions for building programs using Pygame',
    packages=find_packages(),
    include_package_data=True,
    license="BSD",
    url='https://github.com/IrvKalb/pyghelpers',
    install_requires=[
        'pygame-ce>=2.0',
        'pygwidgets>=1.0',        
        ],
    keywords="Timer classes, Scene and SceneMgr classes, Dialogs",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent"
      ]
    )



# High Scores scene
import pygwidgets
import pyghelpers
from HighScoresData import *

def showCustomAnswerDialog(theWindow, theText):
    oDialogBackground = pygwidgets.Image(theWindow, (35, 450),
                                                'images/dialog.png')
    oPromptDisplayText = pygwidgets.DisplayText(theWindow, (0, 480),
                                                theText, width=WINDOW_WIDTH,
                                                justified='center', fontSize=36)
    oUserInputText = pygwidgets.InputText(theWindow, (200, 550), '',
                                                fontSize=36, initialFocus=True)
    oNoButton = pygwidgets.CustomButton(theWindow, (65, 595),
                                                'images/noThanksNormal.png',
                                                over='images/noThanksOver.png',
                                                down='images/noThanksDown.png',
                                                disabled='images/noThanksDisabled.png')
    oYesButton = pygwidgets.CustomButton(theWindow, (330, 595),
                                                'images/addNormal.png',
                                                over='images/addOver.png',
                                                down='images/addDown.png',
                                                disabled='images/addDisabled.png')
    userAnswer = pyghelpers.customAnswerDialog(theWindow,
                                                oDialogBackground,
                                                oPromptDisplayText, oUserInputText,
                                                oYesButton, oNoButton)
    return userAnswer

def showCustomResetDialog(theWindow, theText):
    oDialogBackground = pygwidgets.Image(theWindow,
                                               (35, 450), 'images/dialog.png')
    oPromptDisplayText = pygwidgets.DisplayText(theWindow, (0, 480),
                                                theText, width=WINDOW_WIDTH,
                                                justified='center', fontSize=36)
    oNoButton = pygwidgets.CustomButton(theWindow, (65, 595),
                                                'images/cancelNormal.png',
                                                over='images/cancelOver.png',
                                                down='images/cancelDown.png',
                                                disabled='images/cancelDisabled.png')
    oYesButton = pygwidgets.CustomButton(theWindow, (330, 595),
                                                'images/okNormal.png',
                                                over='images/okOver.png',
                                                down='images/okDown.png',
                                                disabled='images/okDisabled.png')
    choiceAsBoolean = pyghelpers.customYesNoDialog(theWindow,
                                                oDialogBackground, oPromptDisplayText,
                                                oYesButton, oNoButton)
    return choiceAsBoolean


class SceneHighScores(pyghelpers.Scene):
    def __init__(self, window):
        self.window = window
        self.oHighScoresData = HighScoresData()
        
        self.backgroundImage = pygwidgets.Image(self.window,
                                                (0, 0),
                                                'images/highScoresBackground.jpg')

        self.namesField = pygwidgets.DisplayText(self.window, (260, 84), '',
                                                   fontSize=48, textColor=BLACK,
                                                   width=300, justified='left')
        self.scoresField = pygwidgets.DisplayText(self.window,
                                                  (25, 84), '', fontSize=48,
                                                  textColor=BLACK,
                                                  width=175, justified='right')

        self.quitButton = pygwidgets.CustomButton(self.window,
                                                  (30, 650),
                                                  up='images/quitNormal.png',
                                                  down='images/quitDown.png',
                                                  over='images/quitOver.png',
                                                  disabled='images/quitDisabled.png')

        self.backButton = pygwidgets.CustomButton(self.window,
                                                 (240, 650),
                                                 up='images/backNormal.png',
                                                 down='images/backDown.png',
                                                 over='images/backOver.png',
                                                 disabled='images/backDisabled.png')

        self.resetScoresButton = pygwidgets.CustomButton(self.window,
                                                 (450, 650),
                                                 up='images/resetNormal.png',
                                                 down='images/resetDown.png',
                                                 over='images/resetOver.png',
                                                 disabled='images/resetDisabled.png')

        self.showHighScores()

    def getSceneKey(self):
        return SCENE_HIGH_SCORES

    def enter(self, newHighScoreValue=None):
        # This can be called two different ways:
        # 1. If no new high score, newHighScoreValue will be None
        # 2. newHighScoreValue is score of the current game - in top 10
        if newHighScoreValue is None:
            return  # nothing to do

        self.draw() # draw before showing dialog
        # We have a new high score sent in from the Play scene
        dialogQuestion = ('To record your score of ' +
                                 str(newHighScoreValue) + ',\n' +
                                 'please enter your name:')
        playerName = showCustomAnswerDialog(self.window,
                                                                    dialogQuestion)
        if playerName is None:
            return  # user pressed Cancel

        # Add user and score to high scores
        if playerName == '':
            playerName = 'Anonymous'
        self.oHighScoresData.addHighScore(playerName,
                                                            newHighScoreValue)

        # Show the updated high scores table
        self.showHighScores()

    def showHighScores(self):
        # Get the scores and names, show them in two fields
        scoresList, namesList = self.oHighScoresData.getScoresAndNames()
        self.namesField.setValue(namesList)
        self.scoresField.setValue(scoresList)        

    def handleInputs(self, eventsList, keyPressedList):
        for event in eventsList:
            if self.quitButton.handleEvent(event):
                self.quit()

            elif self.backButton.handleEvent(event):
                self.goToScene(SCENE_PLAY)

            elif self.resetScoresButton.handleEvent(event):
                confirmed = showCustomResetDialog(self.window,
                                        'Are you sure you want to \nRESET the high scores?')
                if confirmed:
                    self.oHighScoresData.resetScores()
                    self.showHighScores()

    def draw(self):
        self.backgroundImage.draw()
        self.scoresField.draw()
        self.namesField.draw()
        self.quitButton.draw()
        self.resetScoresButton.draw()
        self.backButton.draw()

    def respond(self, requestID):
        if requestID == HIGH_SCORES_DATA:
            # Request from Play scene for the highest and lowest scores
            # Build a dictionary and return it to the Play scene
            highestScore, lowestScore = self.oHighScoresData.getHighestAndLowest()
            return {'highest':highestScore, 'lowest':lowestScore}



# HighScoresData class
from Constants import *
from pathlib import Path
import json

class HighScoresData():
    """The data file is stored as a list of lists in JSON format.
    Each list is made up of a name and a score:
        [[name, score], [name, score], [name, score] ...]
    In this class, all scores are kept in self.scoresList
    The list is kept in order of scores, highest to lowest.
    """
    def __init__(self):
        self.BLANK_SCORES_LIST = N_HIGH_SCORES * [['-----', 0]]
        self.oFilePath = Path('HighScores.json')

        # Try to open and load the data from the data file
        try:
            data = self.oFilePath.read_text()
        except FileNotFoundError:  # no file, set to blank scores and save
            self.resetScores()
            return

        # File exists, load the scores from the JSON file
        self.scoresList = json.loads(data)

    def addHighScore(self, name, newHighScore):
        # Find the appropriate place to add the new high score
        placeFound = False
        for index, nameScoreList in enumerate(self.scoresList):
            thisScore = nameScoreList[1]
            if newHighScore > thisScore:
                # Insert into proper place, remove last entry
                self.scoresList.insert(index, [name, newHighScore])
                self.scoresList.pop(N_HIGH_SCORES)
                placeFound = True
                break
        if not placeFound:
            return  # score does not belong in the list

        # Save the updated scores
        self.saveScores()

    def saveScores(self):
        scoresAsJson = json.dumps(self.scoresList)
        self.oFilePath.write_text(scoresAsJson)

    def resetScores(self):
        self.scoresList = self.BLANK_SCORES_LIST.copy()
        self.saveScores()

    def getScoresAndNames(self):
        namesList = []
        scoresList = []
        for nameAndScore in self.scoresList:
            thisName = nameAndScore[0]
            thisScore = nameAndScore[1]
            namesList.append(thisName)
            scoresList.append(thisScore)

        return scoresList, namesList

    def getHighestAndLowest(self):
        # Element 0 is highest entry, element -1 is the lowest
        highestEntry = self.scoresList[0]
        lowestEntry = self.scoresList[-1]
        # Get the score (element 1) of each sublist
        highestScore = highestEntry[1]
        lowestScore = lowestEntry[1]
        return highestScore, lowestScore




#
# This is a Blank Scene
#
# This is a template for what a Scene should look like
#

import pygwidgets
import pyghelpers

class MyScene(pyghelpers.Scene):  # Inherits from the Scene class in the pyghelpers file
    def __init__(self, window):
        """
        This method is called when the scene is created
        Create and/or load any assets (images, buttons, sounds)
        that you need for this scene
        """
        self.window = window

        # As a sample, let's create a button
        self.navButton = pygwidgets.TextButton(self.window, (300, 230), 'Navigate')

    def getSceneKey(self):
        return 'Some string or CONSTANT that uniquely represents this scene'

    def enter(self, data):
        """
        This method is called whenever the scene changes to this scene.
        'data' is a any information that is passed from the previous scene
        Typical use is to pass in a dictionary, from which useful data can be extracted.
        """
        pass

    def handleInputs(self, events, keyPressedList):
        """
        This method is called on every frame when an event happens
        It is passed in a list of events and a list of keyboard keys that are down
        Typical code is to loop through the events and handle any that you want to
        This method MUST be included in every scene to override the one in Scene
        """
        for event in events:
            if self.navButton.handleEvent(event):
                print('Clicked on the nav button - typically add a: self.goToScene("NewScene")')

    def update(self):
        """
        This method is called once per frame while the scene is active
        Include in here, any code you want to execute every frame
        """
        pass

    def draw(self):
        """
        This method is called on every frame
        Include any code that you need to draw everything in the window
        (Typical is to do a window fill or show a background image,
        and draw buttons, fields, characters, etc.)
        This method MUST be included in every scene to override the one in Scene
        """
        self.navButton.draw()

    def leave(self):
        """
        This method is called once when your code has asked to move on to a new scene
        It should return any data that this scene wants to pass on to the next scene
        """
        pass



# Constants - used by multiple Python files

WINDOW_WIDTH = 600
WINDOW_HEIGHT = 700
GAME_HEIGHT = 560
DIALOG_BOX_OFFSET = 35
DIALOG_BOX_WIDTH = WINDOW_WIDTH - (2 * DIALOG_BOX_OFFSET)

# Scene keys
SCENE_SPLASH = 'scene splash'
SCENE_PLAY = 'scene play'
SCENE_HIGH_SCORES = 'scene high scores'

WHITE = (255, 255, 255)
BLACK = (0, 0, 0)

POINTS_FOR_GOODIE = 25
POINTS_FOR_BADDIE_EVADED = 1
HIGH_SCORES_DATA = 'high scores data'
N_HIGH_SCORES = 10